

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import pandas as pd

try:
    from adiftools.errors import AdifParserError
except ModuleNotFoundError or ImportError:
    from errors import AdifParserError


def monthly_qso(df, fname):
    ''' plot monthly QSO '''
    if len(df) < 0:
        raise AdifParserError('Empty adif data')

    df['QSO_DATE'] = pd.to_datetime(df['QSO_DATE'])
    df['QSO_DATE'] = df['QSO_DATE'].dt.to_period('M')
    df = df.groupby('QSO_DATE').size().reset_index(name='counts')
    df.set_index('QSO_DATE', inplace=True)

    fig, ax = plt.subplots(figsize=(12, 6))
    bar = ax.bar(df.index.astype(str), df['counts'])

    # basic graph elements
    plt.title('Monthly QSO')
    plt.xlabel('Month')
    plt.ylabel('Number of QSO')
    plt.xticks(rotation=90)
    plt.grid(axis='y')
    ax.set_axisbelow(True)
    plt.legend(['Number of QSO'], loc='upper right')

    # add value on top of each bar
    if len(df) < 24:
        text_rotation = 0
    else:
        text_rotation = 90

    for rect in bar:
        height = rect.get_height()
        ax.text(rect.get_x() + rect.get_width() / 2., height + 1,
                '%d' % int(height), ha='center', va='bottom',
                size='small', rotation=text_rotation)

    # set layout and save to file
    plt.subplots_adjust(left=0.08, right=0.98, bottom=0.18, top=0.93)
    plt.savefig(fname)
    plt.close()


def band_percentage(df, fname):
    ''' generate circle graph for band percentage '''
    # caclulate mode percentage
    if len(df) < 0:
        raise AdifParserError('Empty adif data')

    if 'BAND' in df.columns:
        mode_counts = df['BAND'].value_counts()
    else:
        raise ValueError('BAND column not found in DataFrame')

    # plot circle graph
    _, ax = plt.subplots()
    ax.pie(mode_counts, labels=mode_counts.index, autopct='%1.1f%%',
           startangle=90, counterclock=False)
    plt.title('Band Percentage')
    plt.savefig(fname)
    plt.close()


def monthly_band_qso(df, fname):
    '''バンド別の月間QSO数を積み上げ棒グラフで描画'''
    if len(df) < 0:
        raise AdifParserError('Empty adif data')

    if 'BAND' not in df.columns:
        raise ValueError('BAND column not found in DataFrame')

    from pandas.api.types import PeriodDtype
    if isinstance(df['QSO_DATE'].dtype, PeriodDtype):
        df['QSO_DATE'] = df['QSO_DATE'].dt.to_timestamp()
    else:
        df['QSO_DATE'] = pd.to_datetime(df['QSO_DATE'])
    df['QSO_MONTH'] = df['QSO_DATE'].dt.to_period('M').dt.to_timestamp()
    grouped = df.groupby(['QSO_MONTH', 'BAND']).size().unstack(fill_value=0)
    grouped = grouped.reset_index()
    # grouped['QSO_MONTH'] = grouped['QSO_MONTH'].astype(str)
    grouped['QSO_MONTH'] = grouped['QSO_MONTH'].dt.strftime('%Y-%m')


    # BANDごとの色テーブル（必要に応じて編集）
    band_colors = {
        '2M': '#FF1493',
        '4M': '#CC0044',
        '5M': '#E0E0E0',
        '6M': '#FF0000',
        '8M': '#7F00F1',
        '10M': '#FF69B4',
        '11M': '#00FF00',
        '12M': '#B22222',
        '15M': '#CCA166',
        '17M': '#F2F261',
        '20M': '#F2C40C',
        '30M': '#62D962',
        '40M': '#5959FF',
        '80M': '#E550E5',
        '160M': '#7CFC00',
        # 必要に応じて追加
    }

    fig, ax = plt.subplots(figsize=(12, 6))
    bands = [col for col in grouped.columns if col != 'QSO_MONTH']

    # 未定義BANDにはカラーマップから重複しない色を割り当て
    import itertools
    import matplotlib as mpl
    cmap = plt.get_cmap('tab20')
    used_colors = set(band_colors.values())
    color_cycle = (mpl.colors.to_hex(cmap(i)) for i in range(cmap.N))
    band_color_map = band_colors.copy()
    for band in bands:
        if band not in band_color_map:
            # 未使用色を順に割り当て
            for color in color_cycle:
                if color not in used_colors:
                    band_color_map[band] = color
                    used_colors.add(color)
                    break

    colors = [band_color_map[band] for band in bands]
    grouped.plot(x='QSO_MONTH', y=bands, kind='bar', stacked=True, ax=ax, color=colors)

    plt.title('Monthly QSO by Band')
    plt.xlabel('Month')
    plt.ylabel('Number of QSO')
    plt.xticks(rotation=90)
    plt.grid(axis='y')
    ax.set_axisbelow(True)
    plt.legend(title='Band', loc='upper right')

    # 各バーの上に合計値を表示
    for idx, row in grouped.iterrows():
        total = sum([row[band] for band in bands])
        if len(grouped) < 24:
            text_rotation = 0
        else:
            text_rotation = 90
        ax.text(idx, total + 1, str(int(total)), ha='center', va='bottom', size='small', rotation=text_rotation)

    plt.subplots_adjust(left=0.08, right=0.98, bottom=0.18, top=0.93)
    plt.savefig(fname)
    plt.close()


def main():
    pass


if __name__ == '__main__':
    main()



import pytest
import tempfile
from pathlib import Path
from typing import Any, Generator

from adiftools import adiftools


@pytest.fixture
def prep_instance():
    at = adiftools.ADIFParser()
    file_path = 'tests/sample.adi'
    _ = at.read_adi(file_path)
    return at


@pytest.fixture
def prep_data():
    at = adiftools.ADIFParser()
    file_path = 'tests/sample.adi'
    df = at.read_adi(file_path)
    return df


@pytest.fixture(scope='function')
def txt_file() -> Generator[Path, Any, None]:
    ''' csv tempfile '''
    path = Path(tempfile.NamedTemporaryFile(suffix='.txt', delete=False).name)
    yield path

    # delete tempfile after test
    path.unlink()


def test_read_adi(prep_data):
    ''' test adif DataFrame '''
    assert prep_data.shape == (126, 14)
    assert prep_data.columns.tolist() == [
        'CALL', 'MODE', 'RST_SENT', 'RST_RCVD',
        'QSO_DATE', 'TIME_ON', 'QSO_DATE_OFF',
        'TIME_OFF', 'BAND', 'FREQ', 'STATION_CALLSIGN',
        'MY_GRIDSQUARE', 'COMMENT', 'GRIDSQUARE']


def test_to_adi(prep_instance):
    prep_instance.to_adi('tests/sample_out.adi')
    assert True


def test_plot_monthly(prep_instance):
    prep_instance.plot_monthly('tests/monthly_qso_test.png')
    assert True


def test_plot_band_percentage(prep_instance):
    prep_instance.plot_band_percentage('tests/percentage_band_test.png')
    assert True


def test_number_of_records(prep_instance):
    assert prep_instance.number_of_records == 126


def test_call_to_txt(prep_instance, txt_file):
    prep_instance.call_to_txt(txt_file)
    assert True


@pytest.mark.parametrize(
    # variables
    [
        'data_in',
        'expected_data',
    ],
    # values
    [
        # test cases
        pytest.param('PM85kg', (35.2781423, 136.8735481)),
        pytest.param('PM95pl', (35.4913535, 139.2841430)),
        pytest.param('PM95vq', (35.6812362, 139.7671248)),
        pytest.param('PM53fo', (33.5849988, 130.4490906)),
        pytest.param('PL36te', (26.2001297, 127.6466452)),
        pytest.param('QN00ir', (40.7354587, 140.6904126)),
        pytest.param('QN02us', (42.7791317, 141.6866364)),
        pytest.param('QN01js', (41.7757043, 140.8158222)),
        pytest.param('PM74rs', (34.7861612, 135.4380483)),
        pytest.param('PM63it', (33.8276948, 132.7003773)),
    ]
)
def test_gl2latlon(data_in, expected_data):
    '''Test gridlocator to latitude and longitude conversion.'''

    ERROR_THRESHOLD_COEFFICIENT = 0.55

    coordinates = adiftools.gl_to_latlon(data_in)
    lat_min = expected_data[0] - 1/24 * ERROR_THRESHOLD_COEFFICIENT
    lat_max = expected_data[0] + 1/24 * ERROR_THRESHOLD_COEFFICIENT
    lon_min = expected_data[1] - 1/12 * ERROR_THRESHOLD_COEFFICIENT
    lon_max = expected_data[1] + 1/12 * ERROR_THRESHOLD_COEFFICIENT

    assert (lat_min <= coordinates[0] <= lat_max and
            lon_min <= coordinates[1] <= lon_max)


@pytest.mark.parametrize(
    # variables
    [
        'expected_data',
        'data_in',
    ],
    # values
    [
        # test cases
        pytest.param('PM85kg', (35.2781423, 136.8735481)),
        pytest.param('PM95pl', (35.4913535, 139.2841430)),
        pytest.param('PM95vq', (35.6812362, 139.7671248)),
        pytest.param('PM53fo', (33.5849988, 130.4490906)),
        pytest.param('PL36te', (26.2001297, 127.6466452)),
        pytest.param('QN00ir', (40.7354587, 140.6904126)),
        pytest.param('QN02us', (42.7791317, 141.6866364)),
        pytest.param('QN01js', (41.7757043, 140.8158222)),
        pytest.param('PM74rs', (34.7861612, 135.4380483)),
        pytest.param('PM63it', (33.8276948, 132.7003773)),
        pytest.param('QN01', (41.7757043, 140.8158222, True)),
        pytest.param('PM74', (34.7861612, 135.4380483, True)),
        pytest.param('PM63', (33.8276948, 132.7003773, True)),
        pytest.param('QN01js', (41.7757043, 140.8158222, False)),
        pytest.param('PM74rs', (34.7861612, 135.4380483, False)),
        pytest.param('PM63it', (33.8276948, 132.7003773, False)),
    ]
)
def test_latlon2gl(data_in, expected_data):
    ''' test latitude and longitude to grid locator '''

    if len(data_in) == 3:
        gridlocator = adiftools.latlon2gl(data_in[0], data_in[1], data_in[2])
    else:
        gridlocator = adiftools.latlon2gl(data_in[0], data_in[1])

    assert gridlocator == expected_data


@pytest.mark.parametrize(
    "points, expected",
    [
        ([34.8584, 136.8054, 35.255, 136.9238], 45305.76999847593),
        ([34.8584, 136.8054, 42.7752, 141.6923], 975524.2589462004),
        ([34.8584, 136.8054, 42.2089616, -83.3532049], 10544713.19816745),
    ]
)
def test_get_dist(points, expected):

    TOLERANCE = 0.15

    exp_max = expected + TOLERANCE
    exp_min = expected - TOLERANCE

    res = adiftools.get_dist(points[0], points[1], points[2], points[3])

    assert exp_min < res < exp_max


@pytest.mark.parametrize(
    "callsign, expected",
    [
        ("JA1ABC", True),
        ("7K1XYZ", True),
        ("8L1DEF", True),
        ("JTA1ABC", False),
        ("7Z1XYZ", False),
        ("9J1DEF", False),
    ],
)
def test_is_ja(callsign, expected):
    assert adiftools.is_ja(callsign) == expected


@pytest.mark.parametrize(
    "callsign, expected",
    [
        ("", pytest.raises(ValueError)),
        ("JA", pytest.raises(ValueError)),
        ("7", pytest.raises(ValueError)),
        (7, pytest.raises(TypeError)),

    ],
)
def test_error_is_ja(callsign, expected):
    with expected as e:
        assert adiftools.is_ja(callsign) == e


@pytest.mark.parametrize(
    "callsign, expected",
    [
        ("JS2IIU", 2),
        ("7N4AAA", 1),
        ("JA1RL", 1),
        ("8J1RL", 1),
        ("JR6AAA", 6),
        ("JA0AAA", 0),
        ("JAAAAA", None),
        ("", None),
    ]
)
def test_get_area_num(callsign, expected):
    assert expected == adiftools.get_area(callsign)


# TODO: use test fixture to create a temporary file



import datetime
import re
import multiprocessing as mp

import pandas as pd


try:
    from adiftools.errors import AdifParserError
except ModuleNotFoundError or ImportError:
    from errors import AdifParserError

try:
    from adiftools.adifgraph import monthly_qso, band_percentage
except ModuleNotFoundError or ImportError:
    from adifgraph import monthly_qso, band_percentage

try:
    from adiftools.gridlocator import gl_to_latlon, latlon_to_gl, get_distance
except ModuleNotFoundError or ImportError:
    from gridlocator import gl_to_latlon, latlon_to_gl, get_distance

try:
    from adiftools.callsign import is_ja_call, get_area_num
except ModuleNotFoundError or ImportError:
    from callsign import is_ja_call, get_area_num


class ADIFParser():
    ''' ADIFParser class '''

    def __init__(self):
        ''' initialize ADIFParser class '''
        self._fields = []
        self._number_of_records = 0
        # Pre-compile regex pattern for better performance
        self._adif_pattern = re.compile(r'<(.*?):(\d+)>([^<]*)')
        self.df_adif = pd.DataFrame()

    def read_adi(self, file_path, enable_timestamp=False):
        ''' read adi file and return a DataFrame '''
        records_list = []

        with open(file_path, 'r') as file:
            lines = file.readlines()

        # skip adif header part
        start_line = 0
        for i, line in enumerate(lines):
            if ("<CALL" in line) or ("<call" in line):
                start_line = i
                break

        adif_data = lines[start_line:]

        for record in adif_data:
            record = record.strip()

            if record[:5].upper() == '<CALL' and\
                    record[-5:].upper() == '<EOR>':
                d = self._parse_adif_record(record)
                records_list.append(d)

        # Build DataFrame once from all records
        if records_list:
            df = pd.DataFrame(records_list)
        else:
            df = pd.DataFrame()

        self.df_adif = df
        self._fields = df.columns.tolist()
        self._number_of_records = len(df)

        if len(df) == 0:
            raise AdifParserError('No records found in ADIF file')

        if enable_timestamp:
            # add timestamp column to DataFrame
            df = self._add_timestamp(df)

        return df

    def to_csv(self, file_path):
        ''' save ADIF DataFrame to csv file '''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')
        self.df_adif.to_csv(file_path, index=False)

    def to_excel(self, file_path):
        ''' save ADIF DataFrame to excel file '''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')
        self.df_adif.to_excel(file_path, index=False)

    def read_pickle(self, file_path):
        ''' read DataFrame from pickle file '''
        df = pd.read_pickle(file_path)
        self.df_adif = df
        self._fields = df.columns.tolist()
        self._number_of_records = len(df)

        return df

    def to_pickle(self, file_path):
        ''' save DataFrame to pickle file '''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')
        self.df_adif.to_pickle(file_path)

    def call_to_txt(self, file_path='./call.txt') -> None:
        ''' output callsign in DataFrame to text file '''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')

        calls = set(self.df_adif['CALL'].tolist())

        with open(file_path, 'w') as f:
            for call in calls:
                f.write(f'{call}\n')

    def to_adi(self, file_path):
        ''' save DataFrame to adi file '''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')

        # file extension check
        if file_path[-4:] != '.adi' and file_path[-4:] != '.ADI':
            file_path += '.adi'

        # generate adif header
        header = self._create_adif_header()
        with open(file_path, 'w') as f:
            f.write(header)
            for i in range(len(self.df_adif)):
                record = self.df_adif.iloc[i].to_dict()
                for key, value in record.items():
                    f.write(f'<{key}:{len(str(value))}>{str(value)} ')
                f.write('<EOR>\n')

    @classmethod
    def _add_timestamp(cls, df):
        ''' add timestamp column to DataFrame '''
        if 'QSO_DATE' not in df.columns or 'TIME_ON' not in df.columns:
            raise AdifParserError(
                'QSO_DATE and TIME_ON columns not found in DataFrame')

        df['timestamp'] = pd.to_datetime(
            df['QSO_DATE'] + df['TIME_ON'], format='%Y%m%d%H%M%S')
        return df

    def _parse_adif_record(self, record):
        ''' parse adif record and return a dictionary '''
        fields = self._adif_pattern.findall(record)
        d = {field[0].upper().strip(): field[2].upper().strip()
             for field in fields}
        return d

    def _create_adif_header(self):
        ''' create ADIF header '''
        timestamp = datetime.datetime.now(datetime.UTC)\
            .strftime('%Y%m%d %H%M%S')
        header = f'Generated by adiftools on {timestamp}\n'
        header += f'Total records: {self._number_of_records}\n'
        header += 'visit https://github.com/JS2IIU-MH/adiftools-dev '
        header += 'for more information\n'
        header += f'<CREATED_TIMESTAMP:15>{timestamp}\n'
        header += '<PROGRAMID:9>adiftools\n'
        header += '<EOH>\n\n'
        return header

    @property
    def fields(self):
        return self._fields

    @property
    def number_of_records(self):
        return self._number_of_records

    @property
    def is_loaded(self):
        if len(self.df_adif) > 0:
            return True
        else:
            return False

    # Plot related methods
    def plot_monthly(self, file_path):
        ''' plot monthly QSO bar chart'''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')
        monthly_qso(self.df_adif, file_path)

    def plot_band_percentage(self, file_path):
        ''' plot band percentage pie chart'''
        if len(self.df_adif) == 0:
            raise AdifParserError('No records found in ADIF file')
        band_percentage(self.df_adif, file_path)

    def read_adi_streaming(self, file_path, enable_timestamp=False,
                           chunk_size=1000):
        ''' read adi file using streaming approach for large files '''
        records_list = []
        in_header = True

        with open(file_path, 'r') as file:
            for line in file:
                line = line.strip()

                # Skip header until first CALL record
                if in_header:
                    if ("<CALL" in line) or ("<call" in line):
                        in_header = False
                    else:
                        continue

                # Process ADIF record
                if (line[:5].upper() == '<CALL' and
                        line[-5:].upper() == '<EOR>'):
                    d = self._parse_adif_record(line)
                    records_list.append(d)

                    # Process in chunks to manage memory
                    if len(records_list) >= chunk_size:
                        if not hasattr(self, '_temp_dfs'):
                            self._temp_dfs = []
                        chunk_df = pd.DataFrame(records_list)
                        self._temp_dfs.append(chunk_df)
                        records_list = []

        # Process remaining records
        if records_list:
            if not hasattr(self, '_temp_dfs'):
                self._temp_dfs = []
            chunk_df = pd.DataFrame(records_list)
            self._temp_dfs.append(chunk_df)

        # Combine all chunks
        if hasattr(self, '_temp_dfs') and self._temp_dfs:
            df = pd.concat(self._temp_dfs, ignore_index=True)
            del self._temp_dfs  # Clean up
        else:
            df = pd.DataFrame()

        self.df_adif = df
        self._fields = df.columns.tolist()
        self._number_of_records = len(df)

        if len(df) == 0:
            raise AdifParserError('No records found in ADIF file')

        if enable_timestamp:
            df = self._add_timestamp(df)

        return df

    def read_adi_parallel(self, file_path, enable_timestamp=False,
                          num_processes=None):
        ''' read adi file using parallel processing for very large files '''
        if num_processes is None:
            num_processes = mp.cpu_count()

        # Read all lines first to split work
        with open(file_path, 'r') as file:
            lines = file.readlines()

        # Find start of data
        start_line = 0
        for i, line in enumerate(lines):
            if ("<CALL" in line) or ("<call" in line):
                start_line = i
                break

        adif_data = lines[start_line:]

        # Split data into chunks for parallel processing
        chunk_size = len(adif_data) // num_processes
        chunks = []
        for i in range(0, len(adif_data), chunk_size):
            chunks.append(adif_data[i:i + chunk_size])

        # Process chunks in parallel
        with mp.Pool(processes=num_processes) as pool:
            results = pool.map(self._process_chunk, chunks)

        # Combine results
        all_records = []
        for chunk_records in results:
            all_records.extend(chunk_records)

        # Build final DataFrame
        if all_records:
            df = pd.DataFrame(all_records)
        else:
            df = pd.DataFrame()

        self.df_adif = df
        self._fields = df.columns.tolist()
        self._number_of_records = len(df)

        if len(df) == 0:
            raise AdifParserError('No records found in ADIF file')

        if enable_timestamp:
            df = self._add_timestamp(df)

        return df

    def _process_chunk(self, chunk_lines):
        ''' Process a chunk of lines and return list of parsed records '''
        records = []
        pattern = re.compile(r'<(.*?):(\d+)>([^<]*)')

        for line in chunk_lines:
            line = line.strip()
            if line[:5].upper() == '<CALL' and line[-5:].upper() == '<EOR>':
                fields = pattern.findall(line)
                d = {field[0].upper().strip(): field[2].upper().strip()
                     for field in fields}
                records.append(d)

        return records


# grid locator
def gl2latlon(gridlocator):
    ''' convert grid locator to latitude and longitude in degrees '''
    lat, lon = gl_to_latlon(gridlocator)
    return (lat, lon)


def latlon2gl(latitude, longitude, fourdigit=False):
    ''' convert lat/lon to grid locator '''
    if fourdigit:
        gridlocator = latlon_to_gl(latitude, longitude, fourdigit)
    else:
        gridlocator = latlon_to_gl(latitude, longitude)

    return gridlocator


def get_dist(lat1, lon1, lat2, lo2):
    return get_distance(lat1, lon1, lat2, lo2)


# call sign
def is_ja(call_sign):
    return is_ja_call(call_sign)


def get_area(call_sign):
    return get_area_num(call_sign)


def main():
    file_path = 'tests/sample.adi'
    parser = ADIFParser()
    _ = parser.read_adi(file_path)

    # df.to_csv('tests/sample.csv')
    # print(df.head(50))


if __name__ == '__main__':
    main()
